Function from_outer_reference_column
pub fn from_outer_reference_column(
col: &Column,
schema: &Arc<DFSchema>,
) -> Result<Expression, DataFusionError>Expand description
Convert an outer reference column to a Substrait field reference. Outer reference columns reference columns from an outer query scope in correlated subqueries. We convert them the same way as regular columns since the subquery plan will be reconstructed with the proper schema context during consumption.